/*
 * heat2d-e.c -- solve the heat equation in a two dimensional square region
 * by explicit finite difference method.
 *
 *  cc -o heat2d-e heat2d-e.c -lmatrix -lglscd -lX11 -lm
 */

#include <stdio.h>
#include <math.h>

/* to use matrix, new_matrix() */
#include <matrix.h>

/* to use GLSC */
#define G_DOUBLE
#include <glsc.h>

main()
{
    int Nx, Ny, i, j, k, skip, Kmax;
    double hx, hy, lambdax, lambday, tau, Tmax, t, c, dt;
    matrix u, newu;
    double f();

    printf("2次元正方形領域 (0,1)×(0,1) における熱方程式を陽解法で解く。\n");
    printf("分割数 Nx, Ny: ");
    scanf("%d %d", &Nx, &Ny);

    if ((u = new_matrix(Nx + 1, Ny + 1)) == NULL) {
        fprintf(stderr, "配列 u を確保できませんでした。");
        exit(1);
    }
    if ((newu = new_matrix(Nx + 1, Ny + 1)) == NULL) {
        fprintf(stderr, "配列 newu を確保できませんでした。");
        exit(1);
    }
    hx = 1.0 / Nx;
    hy = 1.0 / Ny;

    printf("時間刻み幅 τ (目安は %g): ",
           1.0 / (2.0 * (1.0 / hx / hx + 1.0 / hy / hy)));
    scanf("%lf", &tau);
    printf("最終時刻 Tmax: ");
    scanf("%lf", &Tmax);

    lambdax = tau / (hx * hx);
    lambday = tau / (hy * hy);
    printf("λx=%g, λy=%g\n", lambdax, lambday);

    printf(" [参考 τ=%g] 図を描く時間間隔 Δt: ", tau);
    scanf("%lf", &dt);
    skip = dt / tau;

    g_init("Meta", 250.0, 160.0);
    g_device(G_BOTH);
    g_def_scale(0, 0.0, 1.0, 0.0, 1.0, 30.0, 70.0, 100.0, 72.0);
    g_def_scale(4, -1.0, 1.0, -1.0, 1.0, 30.0, 30.0, 100.0, 100.0);
    g_def_line(0, G_BLACK, 0, G_LINE_SOLID);
    g_sel_scale(0);

    for (i = 0; i <= Nx; i++)
        for (j = 0; j <= Ny; j++)
            u[i][j] = f(i * hx, j * hy);

    c = 1.0 - 2.0 * (lambdax + lambday);

    g_cls();
    g_hidden2(1.0, 1.0, 0.4, -1.0, 1.0, 5.0, 25.0, 20.0, 20.0, 20.0,
              150.0, 100.0, u, Nx + 1, Ny + 1,
              1, G_SIDE_NONE, 2, 1);

    Kmax = Tmax / tau + 0.5;
    for (k = 1; k <= Kmax; k++) {
        for (i = 1; i < Nx; i++)
            for (j = 1; j < Ny; j++)
                newu[i][j] = c * u[i][j]
                    + lambdax * (u[i + 1][j] + u[i - 1][j])
                    + lambday * (u[i][j + 1] + u[i][j - 1]);
        for (j = 0; j <= Ny; j++)
            newu[0][j] = newu[Nx][j] = 0;
        for (i = 0; i <= Nx; i++)
            newu[i][0] = newu[i][Ny] = 0.0;
        for (i = 0; i <= Nx; i++)
            for (j = 0; j <= Ny; j++)
                u[i][j] = newu[i][j];
        if (k % skip == 0) {
            g_cls();
            g_hidden2(1.0, 1.0, 0.4, -1.0, 1.0, 5.0, 25.0, 20.0, 20.0, 20.0,
                      150.0, 100.0, u, Nx + 1, Ny + 1,
                      1, G_SIDE_NONE, 2, 1);
        }
        t = tau * k;
    }

    /* マウスでクリックされるのを待つ */
    g_sleep(-1.0);
    /* ウィンドウを消す */
    g_term();
    return 0;
}

double f(x, y)
double x, y;
{
    return 16 * x * (1 - x) * y * (1 - y);
}
